/******************************************************************
 * PalmOS includes.
 ******************************************************************/
#include <Common.h>
#include <System/SysAll.h>
#include <System/SysEvtMgr.h>
#include <UI/UIAll.h>

/******************************************************************
 * Application includes.
 ******************************************************************/
#include "callback.h"
#include "resource.h"

#include "Timesheet.h"
#include "Globals.h"
#include "Database.h"

#include "DetailForm.h"
#include "MainForm.h"
#include "UIUtil.h"

/******************************************************************
 * Module internal function prototypes
 ******************************************************************/
void MainFormChangeDate(Int deltaDays, Boolean refreshGUI);
UInt MainFormErasePriorTo(DateType * priorToDate);
void MainFormGotoDay(void);
void MainFormGotoRecord(UInt recIdx);
void MainFormGotoEarliestDay(void);
void MainFormGotoLatestDay(void);
void MainFormGotoWeekday(Byte dayButton, Boolean refreshGUI);
void MainFormInit(void);
void MainFormRefresh(void);
void MainFormNewEntryAttempt(void);
void MainFormTableDrawCell(VoidPtr tablePtr, Word row, Word column, RectanglePtr bounds);
void MainFormTableRowTally(UInt recIdx,
                                  TSDurationRecType* totalDuration,
                                  TSDurationRecType* totalChargeableDuration );
void MainFormTableDraw(void);
void MainFormTableScroll(Byte dir);
void MainFormScrollButtonUpdate(FormPtr formPtr);

/******************************************************************
 * Erase all existing timesheet entry records and day records PRIOR
 * to the specified date.
 ******************************************************************/
UInt
MainFormErasePriorTo(DateType * priorToDate)
{
  UInt idx;
  UInt priorToDays;
  UInt numEntries;
  UInt recDays;
  UInt recDelCount;
  VoidHand hand;
  TSDayRecType *dayPtr;

  /* Convert the prior to date to days */
  priorToDays = DateToDays(*priorToDate);
  recDelCount = 0;
  /* Work until we run out of day records */
  while (DmNumRecords(TSDatabase) > TSFirstDayRecIdx)
    {
      hand = DmQueryRecord(TSDatabase, TSFirstDayRecIdx);
      dayPtr = MemHandleLock(hand);
      /* convert the record date to days */
      recDays = DateToDays(dayPtr->dayDate);
      numEntries = dayPtr->numEntries;
      MemHandleUnlock(hand);
      /* delete this record? */
      if (recDays < priorToDays)
        {
          /* yes, delete day record */
          DmRemoveRecord(TSDatabase, TSFirstDayRecIdx);
          /* Deleting all entries for day, as well as day record itself */
          recDelCount += numEntries + 1;
          /* and delete all day entries */
          for (idx = 0; idx < numEntries; idx++)
            {
              DmRemoveRecord(TSDatabase, TSFirstDayRecIdx);
            }
        }
      else
        {
          /* no, stop searching */
          break;
        }
    }
  /* Check if we've just deleted the timer entry record (if any) */
  if (TSEntryTimer.recIdx > 0)
  {
    if (TSFirstDayRecIdx + recDelCount >= TSEntryTimer.recIdx)
    {
      /* Stop timing as the timed record has been deleted */
      TSEntryTimer.recIdx = 0;
      TSEntryTimer.secs = 0;
    }
    else
    {
      /* Adjust the timed record index accordingly */
      TSEntryTimer.recIdx -= recDelCount;
    }
  }
  return recDelCount;
}

/******************************************************************
 * Event handler for the main form.
 ******************************************************************/
Boolean
MainFormEventHandler(EventPtr event)
{
  DateType priorToDate;
  SWord day;
  SWord month;
  SWord year;
  UInt priorToDays;
  VoidHand hand;
  FormPtr formPtr;
  Char priorTo[20 + 1];
  Boolean handled = false;

  if (event->eType == ctlSelectEvent)
    {
      /* a control button was clicked */
      if (event->data.ctlSelect.controlID >= buttonID_tsDay0 && event->data.ctlSelect.controlID <= buttonID_tsDay6)
        {
          /* one of the day buttons was selected */
          MainFormGotoWeekday(event->data.ctlSelect.controlID - buttonID_tsDay0, true);
          handled = true;
        }
      else
        {
          /* some other button or control was selected */
          if (event->data.ctlSelect.controlID == buttonID_tsNextWeek)
            {
              /* select the next week */
              MainFormChangeDate(7, true);
              handled = true;
            }
          else if (event->data.ctlSelect.controlID == buttonID_tsPrevWeek)
            {
              /* select the previous week */
              MainFormChangeDate(-7, true);
              handled = true;
            }
          else if (event->data.ctlSelect.controlID == buttonID_tsNew)
            {
              /* attempt create a new day entry  */
              MainFormNewEntryAttempt();
              handled = true;
            }
          else if (event->data.ctlSelect.controlID == buttonID_tsGoto)
            {
              /* jump to user specified date */
              day = TSPresentDayDate.day;
              month = TSPresentDayDate.month;
              year = TSPresentDayDate.year + 1920;
              hand = DmGet1Resource('tSTR', stringID_tsSelectDate);
              if (SelectDay(selectDayByDay, &month, &day, &year, (CharPtr) MemHandleLock(hand)))
                {
                  /* user has entered date to go to, save the new date */
                  TSPresentDayDate.day = day;
                  TSPresentDayDate.month = month;
                  TSPresentDayDate.year = year - 1920;
                  MainFormGotoDay();
                  MainFormTableDraw();
                }
              MemHandleUnlock(hand);
              DmReleaseResource(hand);
              handled = true;
            }
          else if (event->data.ctlSelect.controlID == buttonID_tsGotoTimed)
            {
              /* User wants to go to timed entry in database */
              if (TSEntryTimer.recIdx > 0 && TSEntryTimer.secs > 0)
              {
                /* Go to the timed entry */
                MainFormGotoRecord(TSEntryTimer.recIdx);
              }
              else
              {
                /* No entry being timed */
                FrmAlert(alertID_tsNoTimedEntry);
              }
            }
          else if (event->data.ctlSelect.controlID == buttonID_tsWeekViewButton)
            {
              /* Change to the weekly summary view */
              FrmGotoForm(formID_tsWeekly);
              handled = true;
            }
        }
    }
  else if (event->eType == ctlRepeatEvent)
    {
      if (event->data.ctlSelect.controlID == buttonID_tsTableUp)
        {
          /* scroll table up */
          MainFormTableScroll(pageUpChr);
          MainFormTableDraw();
        }
      else if (event->data.ctlSelect.controlID == buttonID_tsTableDown)
        {
          /* scroll table down */
          MainFormTableScroll(pageDownChr);
          MainFormTableDraw();
        }
    }
  else if (event->eType == menuEvent)
    {
      /* a menu item has been selected */
      if (event->data.menu.itemID == menuitemID_tsGotoEarliest)
        {
          /* user wants to go to earliest day in database */
          MainFormGotoEarliestDay();
        }
      else if (event->data.menu.itemID == menuitemID_tsGotoLatest)
        {
          /* user wants to go to latest day in database */
          MainFormGotoLatestDay();
        }
      else if (event->data.menu.itemID == menuitemID_tsEraseDay)
        {
          /* user wants to erase all entries in the current day */
          if (TSPresentDayRecExists)
            {
              /* there's at least one entry that can be deleted for today */
              hand = DmGet1Resource('tSTR', stringID_tsDay);
              if (FrmCustomAlert(alertID_tsConfirmDayWeekDel,
                  (CharPtr)MemHandleLock(hand), NULL, NULL) == 0)
                {
                  /* user confirms delete, delete all entries for today */
                  while (DetailFormDeleteEntry(TSPresentDayRecIdx + 1) != 0)
                    {
                      /* body intentionally empty */
                    }
                  /* nothing to display, move table to top row */
                  TSTableTopEntry = 0;
                  /* Main form has changed */
                  MainFormRefresh();
                }
              MemHandleUnlock(hand);
              DmReleaseResource(hand);
            }
          else
            {
              /* nothing to do! no entries for today */
              hand = DmGet1Resource('tSTR', stringID_tsDay);
              FrmCustomAlert(alertID_tsEmptyDayWeek, (CharPtr)MemHandleLock(hand), NULL, NULL);
              MemHandleUnlock(hand);
              DmReleaseResource(hand);
            }
          handled = true;
        }
      else if (event->data.menu.itemID == menuitemID_tsErasePrior)
        {
          /* user wants to erase all entries prior to a certain date */
          day = TSPresentDayDate.day;
          month = TSPresentDayDate.month;
          year = TSPresentDayDate.year + 1920;
          hand = DmGet1Resource('tSTR', stringID_tsErasePriorTo);
          if (SelectDay(selectDayByDay, &month, &day, &year, (CharPtr) MemHandleLock(hand)))
            {
              /* user has entered date to purge prior to, make it so... */
              DateToDOWDMFormat(month, day, year,
                                (TSAppPrefs.prefFlags & TSPrefShortDateFlag) ? TSSysDateFormat : TSSysLongDateFormat,
                                priorTo);
              if (FrmCustomAlert(alertID_tsConfirmPrior, priorTo, NULL, NULL) == 0)
                {
                  /* user has confirmed prior to erase...do it */
                  priorToDate.day = day;
                  priorToDate.month = month;
                  priorToDate.year = year - 1920;
                  priorToDays = DateToDays(priorToDate);
                  if (MainFormErasePriorTo(&priorToDate) > 0)
                    {
                      /* have actually purged some records, must update globals and form */
                      FindDayRec(&TSPresentDayDate, &TSPresentDayRecIdx, &TSPresentDayRecExists);
                      /* check if we need to re-draw main table */
                      if (priorToDays >= DateToDays(TSPresentDayDate))
                        {
                          /* the current day has been erased, move table to top row */
                          TSTableTopEntry = 0;
                          MainFormRefresh();
                        }
                    }
                }
            }
          MemHandleUnlock(hand);
          DmReleaseResource(hand);
          handled = true;
        }
      else if (event->data.menu.itemID == menuitemID_tsPref)
        {
          /* display the prefs form */
          FrmPopupForm(formID_tsPref);
          handled = true;
        }
      else if (event->data.menu.itemID == menuitemID_tsAbout)
        {
          /* display the About form */
          FrmPopupForm(formID_tsAbout);
          /* FrmHelp(helpID_tsAbout); */
          handled = true;
        }
    }
  else if (event->eType == tblEnterEvent)
    {
      /* a row in the main table has been selected, note I've used tblEnterEvents instead of tblSelectEvents so the
       * user doesn't get to see the annoying GUI 'item inversion' behaviour, which just looks plain icky. */
      SndPlaySystemSound(sndClick);
      /* set the record index of the entry we want to edit, depending on what record currently appears at the
       * top of the table. */
      TSEditEntryIdx = TSPresentDayRecIdx + TSTableTopEntry + event->data.tblSelect.row + 1;
      /* show the details form to the user */
      FrmPopupForm(formID_tsDetail);
      handled = true;
    }
  else if (event->eType == keyDownEvent)
    {
      /* key has been pressed, is it one app should respond to? */
      if (event->data.keyDown.chr == pageUpChr)
        {
          /* command key up, goto previous day */
          /* go back a day */
          MainFormChangeDate(-1, true);
          handled = true;
        }
      else if (event->data.keyDown.chr == pageDownChr)
        {
          /* command key down, goto next day */
          MainFormChangeDate(1, true);
          handled = true;
        }
    }
  else if (event->eType == frmGotoEvent)
    {
      formPtr = FrmGetActiveForm();
      /* initialise the main form to today */
      MainFormInit();
      /* draw the main form */
      FrmDrawForm(formPtr);
      MainFormGotoDay();
      MainFormRefresh();
      /* Go to a specific record on the main form, req'd for Find functionality */
      MainFormGotoRecord(event->data.frmGoto.recordNum);
      handled = true;
    }
  else if (event->eType == frmOpenEvent)
    {
      /* the Main form has been opened, this will only occur once
       * each time the application is opened, so it's safe to
       * initialise the main form here. */
      formPtr = FrmGetActiveForm();
      /* initialise the main form to today */
      MainFormInit();
      /* draw the main form */
      FrmDrawForm(formPtr);
      MainFormGotoDay();
      MainFormRefresh();
      handled = true;
    }
  return handled;
}

/******************************************************************
 * Go to the current date, modifying application internal variables
 * as well as the Main Form controls appropriately.
 ******************************************************************/
void
MainFormGotoDay(void)
{
  int dayButton;
  int weekDay;

  /* set the appropriate main form day button for today */
  weekDay = DayOfWeek(TSPresentDayDate.month, TSPresentDayDate.day, TSPresentDayDate.year + 1920);
  if (TSSysWeekStartDay == 0)
    {
      /* week starts Sunday */
      dayButton = weekDay;
    }
  else
    {
      /* week starts Monday */
      dayButton = (weekDay == 0) ? (6) : (weekDay - 1);
    }
  FrmSetControlGroupSelection(FrmGetActiveForm(), groupID_tsDayGroup, buttonID_tsDay0 + dayButton);
  TSSelectedDayButton = dayButton;
  /* find the day record for today (if any) */
  FindDayRec(&TSPresentDayDate, &TSPresentDayRecIdx, &TSPresentDayRecExists);
  /* refresh the main form date */
  FormUpdateDay(&TSPresentDayDate);
}

/******************************************************************
 *
 ******************************************************************/
void
MainFormGotoRecord(UInt recIdx)
{
  VoidHand recHand;
  TSDayRecType *dayPtr;
  UInt dayRecIdx;

  /* start fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_PROLOGUE;

  /* Have a record number, need to find day record that contains entry record number. */
  dayRecIdx = FindPrevDayRecIdx(TSDatabase, recIdx);
  recHand = DmQueryRecord(TSDatabase, dayRecIdx);
  dayPtr = MemHandleLock(recHand);
  TSPresentDayDate.day = dayPtr->dayDate.day;
  TSPresentDayDate.month = dayPtr->dayDate.month;
  TSPresentDayDate.year = dayPtr->dayDate.year;
  MemHandleUnlock(recHand);
  /* Refresh display */
  MainFormGotoDay();
  /* Open the specified entry */
  TSEditEntryIdx = recIdx;
  /* show the details form to the user */
  FrmPopupForm(formID_tsDetail);

  /* end fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_EPILOGUE;
}

/******************************************************************
 * User has clicked one of the week-day buttons on the main form to
 * change the day of the present week.
 ******************************************************************/
void
MainFormGotoEarliestDay(void)
{
  /* Find the earliest day rec (if one exists) */
  if (FindEarliestDayRec())
    {
      /* go to the earliest day, updating GUI and internal globals */
      MainFormGotoDay();
      MainFormTableDraw();
    }
  else
    {
      /* no earlier day to go to, alert user */
      FrmAlert(alertID_tsEmptyDB);
    }
}

/******************************************************************
 * User has clicked one of the week-day buttons on the main form to
 * change the day of the present week.
 ******************************************************************/
void
MainFormGotoLatestDay(void)
{
  if (FindLatestDayRec())
    {
      /* go to the latest day, updating GUI and internal globals */
      MainFormGotoDay();
      MainFormTableDraw();
    }
  else
    {
      /* no later day to go to, alert user */
      FrmAlert(alertID_tsEmptyDB);
    }
}

/******************************************************************
 * User has clicked one of the week-day buttons on the main form to
 * change the day of the present week.
 ******************************************************************/
void
MainFormGotoWeekday(Byte dayButton, Boolean refreshGUI)
{
  Int deltaDays;

  /* check if the day has really changed */
  if (dayButton == TSSelectedDayButton)
    {
      /* nope, do nothing */
      return;
    }
  else
    {
      /* calculate the CHANGE in days depending on the week day button selected */
      deltaDays = dayButton - TSSelectedDayButton;
      /* save the new day button selected */
      TSSelectedDayButton = dayButton;
      /* modify the date accordingly */
      MainFormChangeDate(deltaDays, refreshGUI);
    }
}

/******************************************************************
 * Change the date of the day presently being displayed, will also
 * update the GUI appropriately if refreshGUI is set to true.
 ******************************************************************/
void
MainFormChangeDate(Int deltaDays, Boolean refreshGUI)
{
  Int weekDay;

  /* modify the date accordingly */
  DateAdjust(&TSPresentDayDate, deltaDays);
  /* find the day record for today (if any) */
  FindDayRec(&TSPresentDayDate, &TSPresentDayRecIdx, &TSPresentDayRecExists);
  /* refresh the main form if req'd */
  if (refreshGUI)
    {
      weekDay = DayOfWeek(TSPresentDayDate.month, TSPresentDayDate.day, TSPresentDayDate.year + 1920);
      if (TSSysWeekStartDay == 0)
        {
          /* week starts Sunday */
          TSSelectedDayButton = weekDay;
        }
      else
        {
          /* week starts Monday */
          TSSelectedDayButton = (weekDay == 0) ? (6) : (weekDay - 1);
        }
      FrmSetControlGroupSelection(FrmGetActiveForm(), groupID_tsDayGroup, buttonID_tsDay0 + TSSelectedDayButton);
      FormUpdateDay(&TSPresentDayDate);
      /* reset TSTableTopEntry now we've changed day */
      TSTableTopEntry = 0;
      MainFormTableDraw();
    }
}

/******************************************************************
 * Initialise the main form, which is about to be shown for the first time.
 ******************************************************************/
void
MainFormInit(void)
{
  Char chr;
  Word id;
  CharPtr str1, str2;
  FormPtr formPtr;

  /* If the start-day-of-week is monday rearrange the labels on the days-of-week push buttons.
   * Note: This code is lifted from the PalmPilot Datebook App, no point reinventing the wheel eh? */
  if (TSSysWeekStartDay == monday)
    {
      formPtr = FrmGetActiveForm();
      str1 = CtlGetLabel(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, buttonID_tsDay0)));
      chr = *str1;
      for (id = buttonID_tsDay1; id <= buttonID_tsDay6; id++)
        {
          str2 = CtlGetLabel(FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, id)));
          *str1 = *str2;
          str1 = str2;
        }
      *str1 = chr;
    }
  /* Select the day view button */
  FrmSetControlGroupSelection(FrmGetActiveForm(), groupID_tsViewGroup, buttonID_tsDayViewButton);
  /* initialise the table globals */
  TSTableTopEntry = 0;
}

/******************************************************************
 * Must be called each time the Main form is returned to. Updates the
 * Main Form table and any other appropriate controls.
 ******************************************************************/
void MainFormRefresh(void)
{
  VoidHand recHand;
  BitmapPtr bitmapPtr;
  ControlPtr controlPtr;
  FormPtr formPtr;

  /*** Get the active Main form ***/
  formPtr = FrmGetActiveForm();

  /*** Redraw the table ***/
  MainFormTableDraw();

  /*** Show or hide the timer button as appropriate ***/
  controlPtr = FrmGetObjectPtr( formPtr, FrmGetObjectIndex( formPtr, buttonID_tsGotoTimed ) );
  if ( TSEntryTimer.recIdx > 0 && TSEntryTimer.secs > 0 )
  {
    /*** Timed entry exists, show button ***/
    CtlShowControl( controlPtr );

    /*** Draw the bitmap ***/
    recHand = DmGet1Resource( 'Tbmp', bitmapID_tsStopWatch );
    bitmapPtr = MemHandleLock( recHand );
    controlPtr = FrmGetObjectPtr( formPtr, FrmGetObjectIndex( formPtr, bitmapID_tsStopWatch ) );
    WinDrawBitmap( bitmapPtr, controlPtr->bounds.topLeft.x, controlPtr->bounds.topLeft.y );
    MemHandleUnlock( recHand );
    DmReleaseResource( recHand );
  }
  else
  {
    /*** No time entry exists, hide button ***/
    CtlHideControl( controlPtr );
  }
}

/******************************************************************
 * User attempts to create a new timesheet entry for the present
 * day if possible.
 ******************************************************************/
void
MainFormNewEntryAttempt(void)
{
  VoidHand hnd;
  TSDayRecType *dayPtr;

  /* check if any new entry records can be inserted into the current day record */
  if (TSPresentDayRecExists)
    {
      /* get the record for the present day */
      hnd = DmQueryRecord(TSDatabase, TSPresentDayRecIdx);
      dayPtr = MemHandleLock(hnd);
      /* have a record for the present day, is it full? */
      if (dayPtr->numEntries >= TSMaxEntriesPerDay)
        {
          /* present day is full, tell the user and bail */
          MemHandleUnlock(hnd);
          FrmAlert(alertID_tsDayFull);
          return;
        }
      MemHandleUnlock(hnd);
    }                           /* else don't have a record for today, may have to create one later if the user saves their new entry */
  /* allow the user to enter details for the *new* record */
  TSEditEntryIdx = 0;           /* editing potentially *new* record */
  /* show the details form to the user */
  FrmPopupForm(formID_tsDetail);
}

/******************************************************************
 * Custom cell drawing routine for the MainForm table.
 ******************************************************************/
void
MainFormTableDrawCell(VoidPtr tablePtr, Word row, Word column, RectanglePtr bounds)
{
  Char txt[25 + 1];
  Int idx;
  SWord txtWidth;
  SWord txtLen;
  Boolean txtFitsInWidth;
  FontID prevFont;
  VoidHand listHand;
  TSCatRecType *listPtr;
  RectangleType rowRect =
  {
    {0, 0},
    {159, 11}
  };                 /* warning: note these are hardcoded bounds! */

  /* start fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_PROLOGUE;

  /* get the entry string index from the item */
  idx = TblGetItemInt(tablePtr, row, column);
  /* format the column appropriately (either text or HOURmin) */
  if (column == 0)
    {
      /* erase the entire row once before we draw the first column. This reduces screen flicker, and is more efficient
       * than erasing each individual cell before re-filling it. */
      rowRect.topLeft.x = bounds->topLeft.x;
      rowRect.topLeft.y = bounds->topLeft.y;
      WinEraseRectangle(&rowRect, 0);
      /* draw the hour in the large font */
      prevFont = FntSetFont(boldFont);
      if (idx & 0x0F)
        {
          if ((idx & 0x0F) > 9)
            {
              txt[0] = '1';     /* actually idx is hour/mins */
              txt[1] = 48 + (idx & 0x0F) - 10;
              txt[2] = NULL;
              WinDrawChars(txt, 2, bounds->topLeft.x, bounds->topLeft.y);
            }
          else
            {
              txt[0] = 48 + (idx & 0x0F);       /* actually idx is hour/mins */
              txt[1] = NULL;
              WinDrawChars(txt, 1, bounds->topLeft.x + 6, bounds->topLeft.y);
            }
        }
      /* draw the minutes in the normal font */
      txt[0] = 48 + ((idx & 0xE0) >> 5);
      txt[1] = 48 + (5 * ((idx & 0x10) >> 4));
      txt[2] = NULL;
      FntSetFont(stdFont);
      WinDrawChars(txt, 2, bounds->topLeft.x + 12, bounds->topLeft.y);
    }
  else
    {
      /* has this row been marked as chargeable, by underlining the current cell */
      if (idx < 0)
        {
          /* fixup index */
          idx = (-idx) - 1;
          /* turn on underlining */
          WinSetUnderlineMode(solidUnderline);
        }
      listHand = DmQueryRecord(TSDatabase, column);
      listPtr = MemHandleLock(listHand);

      /* use the resource translation table to convert index to actual entry index */
      idx = listPtr->transTable[idx];

      /* get the string to put in the cell */
      StrCopy(txt, listPtr->cats[idx]);
      MemHandleUnlock(listHand);

      /* trim the client string to fit within the cell */
      prevFont = FntSetFont(stdFont);

      /* draw the cell contents */
      txtWidth = bounds->extent.x;
      txtLen = StrLen( txt );
      FntCharsInWidth( txt, &txtWidth, &txtLen, &txtFitsInWidth );
      txt[ txtLen ] = NULL;
      WinDrawChars( txt, txtLen, bounds->topLeft.x, bounds->topLeft.y );

      /* turn off underlining */
      WinSetUnderlineMode(noUnderline);
    }
  /* restore the previous font */
  FntSetFont(prevFont);

  /* end fixup from the GNU Palmpilot SDK FAQ so callback function can access global variables */
  CALLBACK_EPILOGUE;
}

/******************************************************************
 * Tally up the next entry record in the main table
 ******************************************************************/
void MainFormTableRowTally(UInt recIdx,
                                  TSDurationRecType* totalDuration,
                                  TSDurationRecType* totalChargeableDuration )
{
  VoidHand recHnd;
  TSEntryRecType* recPtr;
  UInt recAttr;
  Int itemInt;

  /* get the attributes and the entry record this row corresponds to */
  DmRecordInfo(TSDatabase, TSPresentDayRecIdx + recIdx, &recAttr, NULL, NULL);
  recHnd = DmQueryRecord(TSDatabase, TSPresentDayRecIdx + recIdx);
  recPtr = MemHandleLock(recHnd);
  /* tally up hours and minutes for daily total */
  itemInt = recPtr->hours;
  totalDuration->mins += ((itemInt & 0xF0) >> 4);
  totalDuration->hours += (itemInt & 0x0F);
  if (totalDuration->mins >= 12)
    {
      totalDuration->mins -= 12;
      totalDuration->hours += 1;
    }
  /* chargeable entry, so tally up hours and minutes for chargeable daily total */
  if (recAttr & 0x0008)
    {
      totalChargeableDuration->mins += ((itemInt & 0xF0) >> 4);
      totalChargeableDuration->hours += (itemInt & 0x0F);
      if (totalChargeableDuration->mins >= 12)
        {
          totalChargeableDuration->mins -= 12;
          totalChargeableDuration->hours += 1;
        }
    }
  MemHandleUnlock(recHnd);
}

/******************************************************************
 * Fill the main form with entries for the present day (if any).
 ******************************************************************/
void
MainFormTableDraw(void)
{
  Boolean dayHasEntries;
  Byte numEntries;
  UInt recIdx;
  UInt recAttr;
  UInt itemInt;
  Word col;
  Word row;
  VoidHand recHnd;
  VoidPtr recPtr;
  FormPtr formPtr;
  ControlPtr controlPtr;
  Char txt[3];
  Byte txtLen;
  TSDurationRecType totalDuration;
  TSDurationRecType totalChargeableDuration;

  /* get a pointer to the main form as it may not be the visible one at the moment */
  formPtr = FrmGetFormPtr(formID_tsMain);
  /* get the table pointer */
  controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, tableID_tsMain));
  /* reset totals */
  totalDuration.hours = totalDuration.mins = 0;
  totalChargeableDuration.hours = totalChargeableDuration.mins = 0;
  /* find the number of entries in the current day */
  if (TSPresentDayRecExists)
    {
      /* get the number of entries from the day record */
      recHnd = DmQueryRecord(TSDatabase, TSPresentDayRecIdx);
      recPtr = MemHandleLock(recHnd);
      numEntries = ((TSDayRecType *) recPtr)->numEntries;
      MemHandleUnlock(recHnd);
      /* go through all the entries prior to the first record that actually appears in the table, building totals */
      for (recIdx = 1; recIdx < TSTableTopEntry + 1; recIdx++)
        {
          MainFormTableRowTally(recIdx, &totalDuration, &totalChargeableDuration );
        }
    }
  else
    {
      /* no entries for today */
      numEntries = 0;
      recIdx = 0;
    }
  dayHasEntries = numEntries > 0;
  /* go through the remaing records, building the rows in the table */
  for (row = 0; row < TSMainFormTableRows; row++, recIdx++)
    {
      if (recIdx > 0 && recIdx <= numEntries)
        {
          MainFormTableRowTally(recIdx, &totalDuration, &totalChargeableDuration );
          /* get the attributes of the entry record this row corresponds to */
          DmRecordInfo(TSDatabase, TSPresentDayRecIdx + recIdx, &recAttr, NULL, NULL);
          /* get the entry record this row corresponds to */
          recHnd = DmQueryRecord(TSDatabase, TSPresentDayRecIdx + recIdx);
          recPtr = MemHandleLock(recHnd);
          /* fill the row items with record data */
          /* set hour/minute/chargeable column */
          TblSetItemStyle((TablePtr) controlPtr, row, 0, customTableItem);
          TblSetItemInt((TablePtr) controlPtr, row, 0, ((TSEntryRecType *) recPtr)->hours);

          /* set client column */
          TblSetItemStyle((TablePtr) controlPtr, row, 1, customTableItem);
          /* translate the client entry index into the actual client index using the client trans table */
          if ((recAttr & 0x0008) && (TSAppPrefs.prefFlags & TSPrefUlineClientFlag))
            {
              /* entry is chargeable, -ve client index indicates chargeable rows */
              itemInt = -(((TSEntryRecType *) recPtr)->clientIdx + 1);  /* have to add one as idx 0 = 'none', so how do we make
                                                                         * -0 indicate underlining? */
            }
          else
            {
              /* entry is not chargeable, +ve client index indicates non-chargeable rows */
              itemInt = ((TSEntryRecType *) recPtr)->clientIdx;
            }
          TblSetItemInt((TablePtr) controlPtr, row, 1, itemInt);

          /* set project column */
          TblSetItemStyle((TablePtr) controlPtr, row, 2, customTableItem);
          /* translate the client entry index into the actual client index using the client trans table */
          TblSetItemInt((TablePtr) controlPtr, row, 2, ((TSEntryRecType *) recPtr)->projectIdx);

          /* set task column */
          TblSetItemStyle((TablePtr) controlPtr, row, 3, customTableItem);
          /* translate the client entry index into the actual client index using the client trans table */
          TblSetItemInt((TablePtr) controlPtr, row, 3, ((TSEntryRecType *) recPtr)->taskIdx);

          /* finished setting up columns */
          MemHandleUnlock(recHnd);
          /* data for this row, make it usable and selectable */
          TblSetRowUsable((TablePtr) controlPtr, row, true);
          /* mark the row invalid, so it re-draws */
          TblMarkRowInvalid((TablePtr) controlPtr, row);
        }
      else
        {
          /* no data for this row, make it unusable */
          TblSetRowUsable((TablePtr) controlPtr, row, false);
        }
    }
  /* total the remaining entries (if any) */
  for (; recIdx <= numEntries; recIdx++)
    {
      MainFormTableRowTally(recIdx, &totalDuration, &totalChargeableDuration );
    }
  /* make all columns in the table usable */
  for (col = 0; col < TSMainFormTableCols; col += 1)
    {
      TblSetColumnUsable((TablePtr) controlPtr, col, true);
    }
  /* set column spacing */
  TblSetColumnSpacing((TablePtr) controlPtr, 0, 3);
  TblSetColumnSpacing((TablePtr) controlPtr, 1, 3);
  TblSetColumnSpacing((TablePtr) controlPtr, 2, 3);
  TblSetColumnSpacing((TablePtr) controlPtr, 3, 0);
  /* set the custom drawn columns */
  TblSetCustomDrawProcedure((TablePtr) controlPtr, 0, MainFormTableDrawCell);   /* hour / mins column */
  TblSetCustomDrawProcedure((TablePtr) controlPtr, 1, MainFormTableDrawCell);   /* client column */
  TblSetCustomDrawProcedure((TablePtr) controlPtr, 2, MainFormTableDrawCell);   /* project column */
  TblSetCustomDrawProcedure((TablePtr) controlPtr, 3, MainFormTableDrawCell);   /* task column */
  /* redraw the table */
  TblDrawTable((TablePtr) controlPtr);
  if (dayHasEntries)
    {
      /* draw a little line to indicate it's a total */
      controlPtr = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, labelID_tsTotal));
      col = controlPtr->bounds.topLeft.x;
      row = controlPtr->bounds.topLeft.y;
      WinDrawLine(col, row - 1, col + 22, row - 1);

      /* draw the hour and minute totals if req'd */
      if (totalDuration.hours > 0)
        {
          if (totalDuration.hours < 10)
            {
              /* hopefully less code than calling StrIToA? */
              txt[0] = 48 + totalDuration.hours;
              txt[1] = NULL;
              txtLen = 1;
              col += 6;
            }
          else
            {
              StrIToA(txt, totalDuration.hours);
              txtLen = StrLen(txt);
            }
          FntSetFont(boldFont);
          WinDrawChars(txt, txtLen, col, row);
        }
      /* draw the minutes in the normal font */
      FntSetFont(stdFont);
      txt[0] = 48 + ((totalDuration.mins & 0x0E) >> 1);
      txt[1] = 48 + (5 * (totalDuration.mins & 0x01));
      txt[2] = NULL;
      WinDrawChars(txt, 2, controlPtr->bounds.topLeft.x + 12, row);

      /* Draw the chargeable hour and minute totals if req'd.
       * Note, this preference is 'inverted', 1 = Off, 0 = On so it's on by default */
      if (!(TSAppPrefs.prefFlags & TSPrefChargeTotalsFlag))
        {
          col = controlPtr->bounds.topLeft.x + 32;
          row = controlPtr->bounds.topLeft.y;
          WinDrawLine(col, row - 1, col + 22, row - 1);
          if (totalChargeableDuration.hours > 0)
            {
              if (totalChargeableDuration.hours < 10)
                {
                  /* hopefully less code than calling StrIToA? */
                  txt[0] = 48 + totalChargeableDuration.hours;
                  txt[1] = NULL;
                  txtLen = 1;
                  col += 6;
                }
              else
                {
                  StrIToA(txt, totalChargeableDuration.hours);
                  txtLen = StrLen(txt);
                }
              FntSetFont(boldFont);
              WinDrawChars(txt, txtLen, col, row);
            }
          /* draw the minutes in the normal font */
          FntSetFont(stdFont);
          txt[0] = 48 + ((totalChargeableDuration.mins & 0x0E) >> 1);
          txt[1] = 48 + (5 * (totalChargeableDuration.mins & 0x01));
          txt[2] = NULL;
          WinDrawChars(txt, 2, controlPtr->bounds.topLeft.x + 32 + 12, row);
       }
    }
  /* Finally update the on-screen scroll buttons */
  MainFormScrollButtonUpdate(formPtr);
}

/******************************************************************
 * Scroll the table in the specified direction (pageUpChr or pageDownChr)
 ******************************************************************/
void
MainFormTableScroll(Byte dir)
{
  Word numEntries;
  VoidHand recHand;
  TSDayRecType *recPtr;

  /* Find the number of entries in the current day (if any) */
  if (TSPresentDayRecExists)
    {
      recHand = DmQueryRecord(TSDatabase, TSPresentDayRecIdx);
      recPtr = MemHandleLock(recHand);
      numEntries = recPtr->numEntries;
      MemHandleUnlock(recHand);
    }
  else
    {
      numEntries = 0;
    }
  TableScroll(&TSTableTopEntry, TSMainFormTableRows, numEntries, dir);
}

/******************************************************************
 * Update the on-screen scroll up/down buttons.
 ******************************************************************/
void
MainFormScrollButtonUpdate(FormPtr formPtr)
{
  Byte numEntries;
  VoidHand recHand;
  TSDayRecType *recPtr;

  /* Find the number of entries in the current day (if any) */
  if (TSPresentDayRecExists)
    {
      recHand = DmQueryRecord(TSDatabase, TSPresentDayRecIdx);
      recPtr = MemHandleLock(recHand);
      numEntries = recPtr->numEntries;
      MemHandleUnlock(recHand);
    }
  else
    {
      numEntries = 0;
    }
  /* Update the scroll buttons */
  ScrollButtonUpdate(formPtr, TSTableTopEntry, TSMainFormTableRows, numEntries);
}

/*** End of file ***/